/*************************************************************************************************
*
*
*			MacZoop - "the framework for the rest of us"		 
*
*
*
*			ZGWorld.cpp			-- an object for managing an offscreen GWorld
*
*
*
*
*
*			 1996, Graham Cox
*
*
*
*
*************************************************************************************************/


#include	"ZGWorld.h"
#include	"MacZoop.h"
#include	<FixMath.h>

CLASSCONSTRUCTOR( ZGWorld );


/***********************************************************************************************
*
*	CONSTRUCTORS
*
***********************************************************************************************/

// make a GWorld with the parameters passed

ZGWorld::ZGWorld( const Rect& aSize,
				  const short aDepth,
				  const CTabHandle aCTable,
				  const Boolean makeTemp)
	: ZObject()
{
	classID = CLASS_ZGWorld;
	
	gw = NULL;
	pState = 0;
	lockLevel = 0;
	inTempMem = makeTemp;
	
	MakeMacGWorld(aSize, aDepth, aCTable, makeTemp);
	Clear();
}

// make a GWorld from a picture handle

ZGWorld::ZGWorld( const PicHandle aPicture, const Boolean makeTemp )
	: ZObject()
{
	classID = CLASS_ZGWorld;
	
	gw = NULL;
	pState = 0;
	lockLevel = 0;
	inTempMem = makeTemp;
	
	FailNIL( aPicture );
	MakeGWorldFromPicture( aPicture, makeTemp );
}

// make a GWorld from the PICT resource with the ID passed

ZGWorld::ZGWorld( const short pictID, const Boolean makeTemp )
	: ZObject()
{
	classID = CLASS_ZGWorld;

	gw = NULL;
	pState = 0;
	lockLevel = 0;
	inTempMem = makeTemp;
	
	PicHandle	p;
	
	FailNILRes( p = GetPicture( pictID ));
	
	try
	{
		MakeGWorldFromPicture( p, makeTemp );
	}
	catch(OSErr err)
	{
		ReleaseResource((Handle) p );
	
		throw err;
	}
	
	ReleaseResource((Handle) p );
}

// make a ZGWorld from another one

ZGWorld::ZGWorld( ZGWorld* aGW, const Boolean makeTemp )
	: ZObject()
{
	classID = CLASS_ZGWorld;

	FailNILParam( aGW );
	
	Rect		aSize;
	short		aDepth; 
	CTabHandle	aCTable;

	gw = NULL;
	pState = 0;
	lockLevel = 0;
	inTempMem = makeTemp;

	aGW->GetSize( &aSize );
	aDepth = aGW->GetDepth();
	aCTable = aGW->GetColours();
	
	MakeMacGWorld( aSize, aDepth, aCTable, makeTemp );
	Copy( aGW );
}


ZGWorld::ZGWorld()
	: ZObject()
{
	classID = CLASS_ZGWorld;

	gw = NULL;
	pState = 0;
	lockLevel = 0;
	inTempMem = FALSE;
}

/***********************************************************************************************
*
*	DESTRUCTOR
*
***********************************************************************************************/


ZGWorld::~ZGWorld()
{
	// for safety, if the current port or device is part of this object,
	// switch to a valid port and device before disposing
	
	if ( gw )
	{
		CGrafPtr	curPort;
		GDHandle	sDevice;
		
		sDevice = GetGWorldDevice( gw );
		GetPort( &(GrafPtr) curPort );
		
		if ((CGrafPtr) gw == curPort ||
			 sDevice == GetGDevice())
		{ 
			SetGWorld((CGrafPtr) FrontWindow(), GetMainDevice());
		}
	
		DisposeGWorld(gw);
	}
}


/***********************************************************************************************
*
*	LOCK- lock the pixels prior to drawing
*
***********************************************************************************************/


void	ZGWorld::Lock()
{
	// locks the pixels so that you can draw to/from them
	
	if (lockLevel == 0)
	{
		PixMapHandle	pp;
		
		pp = GetGWorldPixMap(gw);
		
		if (! LockPixels(pp))
			FailOSErr(kGWPixelsPurgedErr);
	}
	
	lockLevel++;
}



/***********************************************************************************************
*
*	UNLOCK -frees the pixels after a drawing operation
*
***********************************************************************************************/


void	ZGWorld::Unlock()
{
	// unlocks previously locked pixels
	
	lockLevel--;
	
	if (lockLevel <= 0)
	{
		PixMapHandle	pp;
		
		pp = GetGWorldPixMap(gw);
		UnlockPixels(pp);
		
		lockLevel = 0;
	}
}



/***********************************************************************************************
*
*	SETPURGEABLE_ mark the pixels as purgeable or not
*
***********************************************************************************************/


void	ZGWorld::SetPurgeable( const Boolean aPurgeState )
{
	// sets pixels purgeable or unpurgeable according to state. Note that it is extremely wise to check
	// whether the pixels are really in memory before setting them from purgeable to unpurgeable.
	
	PixMapHandle	pp;
	
	pp = GetGWorldPixMap(gw);
	
	if (aPurgeState)
		AllowPurgePixels(pp);
	else
		NoPurgePixels(pp);
		
	pState = GetPixelsState(pp);
}


/***********************************************************************************************
*
*	GETSTATE- get the current pixels state
*
***********************************************************************************************/

short	ZGWorld::GetState()
{
	return GetPixelsState(GetPixMap());
}


/***********************************************************************************************
*
*	SETCOLOURS- change the colour table of the image, optionally remapping the image
*
***********************************************************************************************/


void	ZGWorld::SetColours( const CTabHandle aCTable, const Boolean reMapColours )
{
	// sets the colour table of the GWorld to the colour table passed in. The table remains the callers.
	// This updates the GWorld, mapping it to the new colours. If you want to substitute a table without
	// mapping the colours, pass a second parameter of FALSE.
	
	if (reMapColours)
	{
		// here we want to re-map the colours to the new table, so we actually completely update the
		// GWorld, warts an' all.
		
		GWorldFlags		gwFlags;
		short			curDepth;
		Rect			curSize;
		
		curDepth = GetDepth();
		GetSize(&curSize);
		
		// set the flags so that the image is preserved
		
		gwFlags = clipPix;
		
		gwFlags = UpdateGWorld(&gw, curDepth, &curSize, aCTable, NULL, gwFlags);
		
		if (gwFlags & gwFlagErr)
			FailOSErr(kGWCTableUpdtFailed);	
	}
	else
	{
		// to change tables without re-mapping, the table is copied over the existing table, and the
		// inverse table is re-built. This does not reallocate the pixels, so false colours may appear.
		// This is presumably what you wanted!
		
		CTabHandle		curTable;
		long			tSize;
		
		curTable = GetColours();
		
		if (curTable == aCTable)
			FailOSErr(paramErr);
			
		FailNIL(aCTable);
		
		// size the current table to the same as the passed one
		
		SetHandleSize((Handle) curTable, tSize = GetHandleSize((Handle) aCTable));
		FailOSErr(MemError());

		// copy the table over
		
		BlockMoveData(*aCTable, *curTable, tSize);
	
		// force the inverse-table to get rebuilt next time
		
		GDeviceChanged(GetGWorldDevice(gw));
	}
}



/***********************************************************************************************
*
*	SETDEPTH- change the bitdepth of the offscreen world
*
***********************************************************************************************/


void	ZGWorld::SetDepth( const short aDepth, const CTabHandle aCTable, const Boolean dither)
{
	// sets the GWorld's depth to that passed, also optionally setting the colour table. If the colour table
	// is NULL, the default for the depth is used. If you only want to change the colour table, use SetColours.
	// This preserves the pixels by clipping.
	
	if (aDepth != GetDepth())
	{
		GWorldFlags		gwFlags;
		Rect			curSize;
		
		GetSize(&curSize);
		
		gwFlags = clipPix;
		
		if (dither)
			gwFlags |= ditherPix;
		
		gwFlags = UpdateGWorld(&gw, aDepth, &curSize, aCTable, NULL, gwFlags);
	
		if (gwFlags & gwFlagErr)
			FailOSErr(kGWChangeDepthFailed);
	
	}
}



/***********************************************************************************************
*
*	SETSIZE- change the bounds rect of the image, optionally rescaling the image
*
***********************************************************************************************/


void	ZGWorld::SetSize( const Rect& aSize, const Boolean scaleImage)
{
	// sets the GWorlds bounds to that passed. The depth and colours remain the same. Normally, the image is
	// preserved by scaling to the new size. Pass FALSE to clip to the new size instead.
	
	Rect			curSize;
	GWorldFlags		gwFlags;
	short			curDepth;
	CTabHandle		curCTab;
	
	GetSize(&curSize);
	
	if (! EqualRect( &aSize, &curSize))
	{
		curDepth = GetDepth();
		curCTab = GetColours();

		if (scaleImage)
			gwFlags = stretchPix;
		else
			gwFlags = clipPix;
			
		gwFlags = UpdateGWorld(&gw, curDepth, &aSize, curCTab, NULL, gwFlags);	
	
		if (gwFlags & gwFlagErr)
			FailOSErr(kGWChangeSizeFailed);
			
		// if the image was cropped and the GWorld is now larger than it was, we need to fill
		// the excess area with white, otherwise it looks orrible.
		
		if (! scaleImage)
		{
			if ((aSize.right > curSize.right) 	||
				(aSize.bottom > curSize.bottom))
			{
				RgnHandle	old, xs;
				CGrafPtr	savePort;
				GDHandle	saveDevice;
				
				Lock();
				
				RectRgn(old = NewRgn(), &curSize);
				RectRgn(xs = NewRgn(), &aSize);
				
				DiffRgn(xs, old, xs);
				
				SetPortToGW(&savePort, &saveDevice);
				EraseRgn(xs);
				SetGWorld(savePort, saveDevice);
				
				Unlock();
				
				DisposeRgn(old);
				DisposeRgn(xs);
			}
		}
	}
}


/***********************************************************************************************
*
*	GETDEPTH- get the current pixel depth of the image
*
***********************************************************************************************/


short	ZGWorld::GetDepth()
{
	PixMapHandle	pp;
	
	pp = GetGWorldPixMap(gw);
	return (*pp)->pixelSize;
}



/***********************************************************************************************
*
*	GETCOLOURS- get the current colour table of the image (NOT A COPY)
*
***********************************************************************************************/


CTabHandle		ZGWorld::GetColours()
{
	// returns a handle to the colour table. Note- this handle belongs to this object, so
	// do not dispose of it. If you modify it, call CTabChanged.
	
	PixMapHandle	pp;
	
	pp = GetGWorldPixMap(gw);
	return (*pp)->pmTable;
}



/***********************************************************************************************
*
*	GETCOLOURINFO- returns the ctSeed field of the current colour table. Useful, honest!
*
***********************************************************************************************/

short	ZGWorld::GetColourInfo()
{
	CTabHandle	aCT = GetColours();

	if (aCT)
		return (*aCT)->ctSeed;
	else
		return -1;
}


/***********************************************************************************************
*
*	GETPIXMAP- get the pixmap internal to the offscreen world. Use with care.
*
***********************************************************************************************/

PixMapHandle	ZGWorld::GetPixMap(Boolean lock)
{
	// returns the GWorld's pixmap, optionally locking it first. If you want to copy to and from
	// the GWorld, use the copy methods instead.
	
	if (lock)
		Lock();
		
	return GetGWorldPixMap(gw);
}


/***********************************************************************************************
*
*	REALLOCPIXMAP- forces a purged GWorld to reallocate the pixmap. If in memory, does nothing.
*
***********************************************************************************************/

void		ZGWorld::ReallocPixMap()
{
	if (! InMemory())
	{
		GWorldFlags		gwFlags;
		short			aDepth, attempts = 0;
		Rect			aSize;
		CTabHandle		aColours;
		Boolean			mSuccess = FALSE;
		long			mSize;
		PixMapHandle	pp;
		
		aDepth = GetDepth();
		GetSize(&aSize);
		aColours = GetColours();
		
		do
		{
			attempts++;
			
			gwFlags = UpdateGWorld(&gw, aDepth, &aSize, aColours, NULL, 0);
		
			if (gwFlags & gwFlagErr)
			{
				// the attempt failed. try various things to get ourselves the memory.
				
				pp = GetGWorldPixMap(gw);
				mSize = ((aSize.bottom - aSize.top) * ((*pp)->rowBytes & 0x3FFF)) + 32768L;
				
				switch (attempts)
				{
					case 1:
						PurgeMem(mSize);	
						break;
					case 2:
						PurgeMem(mSize);
						CompactMem(mSize);
						break;
					case 3:
						MaxMem(&mSize);
						break;
					default:
						FailOSErr(kGWNoMemForPixels);

				}
			}
			else
				mSuccess = TRUE;
		}
		while(! mSuccess);
	}
}


/***********************************************************************************************
*
*	INMEMORY- returns whether pixel buffer has been purged or not
*
***********************************************************************************************/

Boolean		ZGWorld::InMemory()
{
	// returns whether the pixmap is in memory or not. It will return false if the pixels have
	// been purged. This works by trying to lock the pixels.
	
	PixMapHandle	pp;
	Boolean			result;
	
	pp = GetGWorldPixMap(gw);
	pState = GetPixelsState(pp);
	
	result = LockPixels(pp);
	SetPixelsState(pp, pState);
	
	return result;
}


/***********************************************************************************************
*
*	GETSIZE- return the current boundsrect of the image
*
***********************************************************************************************/


void	ZGWorld::GetSize(Rect* aFrame)
{
	// returns the GWorld's bounds rect
	
	*aFrame = gw->portRect;
}


/***********************************************************************************************
*
*	GETIMAGEMEMSIZE- return the memory used by the image buffer. If not in mem, returns 0.
*
***********************************************************************************************/


long	ZGWorld::GetImageMemSize()
{
	long	aSize = 0;
	
	if (InMemory())
	{
		PixMapHandle pp = GetPixMap();
		Rect	br = (*pp)->bounds;
		
		aSize = (long)(br.bottom - br.top) * (long)((*pp)->rowBytes & 0x3FFF);	
	}
	
	return aSize;
}



/***********************************************************************************************
*
*	COPYIN- Copy from a bitmap or pixmap INTO this GWorld.
*
***********************************************************************************************/


void	ZGWorld::CopyIn(BitMap* srcBits, Rect* src, Rect* dest, short mode, RgnHandle mask)
{
	// This method is a single call to copy from a bitmap or pixmap (or other GWorld that is not
	// managed by a wrapper like this) to this GWorld. This saves the port, locks the GWorld,
	// then does a CopyBits with the parameters passed.
	
	CGrafPtr		savePort;
	GDHandle		saveDevice;
	PixMapHandle	pp;

	Lock();	// may throw exception
	SetPortToGW(&savePort, &saveDevice);
	
	pp = GetGWorldPixMap(gw);
	CopyBits(srcBits, (BitMap*) *pp, src, dest, mode, mask);
	
	Unlock();
	SetGWorld(savePort, saveDevice);
}


/***********************************************************************************************
*
*	COPYOUT- copy from this GWorld to another pixmap or bitmap
*
***********************************************************************************************/



void	ZGWorld::CopyOut(BitMap* destBits, Rect* src, Rect* dest, short mode, RgnHandle mask)
{
	// Copies pixel image data from this object to any other bitmap or pixmap. This assumes that
	// the destination port is correctly set.
	
	PixMapHandle	pp;
	
	Lock(); // may throw exception
	
	pp = GetGWorldPixMap(gw);
	CopyBits((BitMap*) *pp, destBits, src, dest, mode, mask);
	
	Unlock();	
}


/***********************************************************************************************
*
*	COPY- completely copy an image from another ZGWorld object INTO this one, scaling the image.
*
***********************************************************************************************/


 
void	ZGWorld::Copy(ZGWorld* aGW, short mode)
{
	// copies the image from another object like this into this one. This scales the image from
	// the source to the dest, applying the mode
	
	PixMapHandle	pp;
	Rect			src, dest;
	
	FailNIL(aGW);
	
	pp = aGW->GetPixMap(TRUE);
	
	aGW->GetSize(&src);
	GetSize(&dest);
	
	CopyIn((BitMap*) *pp, &src, &dest, mode);
	
	aGW->Unlock();	
}


/***********************************************************************************************
*
*	COPYIN- copy from another ZGWorld object INTO this one
*
***********************************************************************************************/


void	ZGWorld::CopyIn(ZGWorld* srcGW, Rect* src, Rect* dest, short mode, RgnHandle mask)
{
	// object version of CopyIn
	PixMapHandle	pp;
	
	FailNIL(srcGW);
	
	pp = srcGW->GetPixMap(TRUE);	// locks the source GWorld
	
	CopyIn((BitMap*) *pp, src, dest, mode, mask);
	
	srcGW->Unlock();
}



/***********************************************************************************************
*
*	COPYOUT- copy from this ZGWorld object to another one
*
***********************************************************************************************/


void	ZGWorld::CopyOut( ZGWorld* destGW, Rect* src, Rect* dest, short mode, RgnHandle mask )
{
	// object version of CopyOut. This actually operates by calling the destination's CopyIn method, since that
	// is a more efficient and cunning way to do it!
	
	FailNILParam( destGW );
	destGW->CopyIn( this, src, dest, mode, mask );
}



/***********************************************************************************************
*
*	CLEAR- erase the image completely
*
***********************************************************************************************/

void	ZGWorld::Clear()
{
	CGrafPtr	savePort;
	GDHandle	saveDevice;
	Rect		r;
	
	Lock();
	SetPortToGW( &savePort, &saveDevice );
	GetSize( &r );
	EraseRect( &r );
	Unlock();
	SetGWorld( savePort, saveDevice );
}



/***********************************************************************************************
*
*	MAKEPICTURE- get the GWorld's image as a picture (makes a new picture handle)
*
***********************************************************************************************/

PicHandle	ZGWorld::MakePicture( Rect* toFitRect, RgnHandle withMaskRgn )
{
	// turns the whole GWorld into a picture and returns it. Be prepared for this to fail!
	
	PicHandle	p = NULL;
	CGrafPtr	savePort;
	GDHandle	saveDevice;
	Rect		src, pr;
	short		mode = srcCopy;
	
	GetSize(&pr);
		
	if ( toFitRect == NULL )
		src = pr;
	else
	{
		// we want to make a picture of the size passed in, but which contains the
		// whole image. The image will be scaled to fit, preserving the aspect ratio
		// of the original image. This is useful for making thumbnails.
	
		Scale2Rects( &pr, toFitRect );
		mode |= ditherCopy;
	}
	
	SetPortToGW( &savePort, &saveDevice );
	ForeColor( blackColor );
	BackColor( whiteColor );
	
	try
	{
		FailNIL( p = OpenPicture( &pr ));
	
		// copy the GWorld to itself- this records the CopyBits in the picture
		
		CopyIn( this, &src, &pr, mode, withMaskRgn );	// this call will lock and unlock the GWorld as needed
		ClosePicture();
	}
	catch( OSErr err )
	{
		if ( p )
		{
			ClosePicture();
			KillPicture( p );
			p = NULL;
		}
		Unlock();
		SetGWorld( savePort, saveDevice );
		
		throw err;
	}
	
	Unlock();
	SetGWorld( savePort, saveDevice );

	// check that the picture is well-formed.
	
	pr = (*p)->picFrame;
	
	if ( EmptyRect( &pr ))
	{
		KillPicture( p );
		FailOSErr( kGWBadPictureErr );
	}
	
	if ( GetHandleSize((Handle) p) <= sizeof( Picture ))
	{
		KillPicture( p );
		FailOSErr( kGWBadPictureSizErr );
	}
	
	// only return a good picture, thanx!
	
	return p;
}


/***********************************************************************************************
*
*	COPYPICTURE- draw a picture into the GWorld
*
***********************************************************************************************/

void		ZGWorld::CopyPicture( PicHandle aPicture, Rect* destRect )
{
	// draw the picture into the GWorld. If destRect is NULL, use the picture's frame, offset to
	// 0,0 as the rect. This is a typical case.
	
	Rect		dr;
	CGrafPtr	savePort;
	GDHandle	saveDevice;
	
	if ( destRect == NULL )
	{
		dr = (*aPicture)->picFrame;
		OffsetRect( &dr, -dr.left, -dr.top );
	}
	else
		dr = *destRect;
		
	// lock the GWorld and draw the picture
	
	Lock();
	SetPortToGW( &savePort, &saveDevice );
	
	DrawPicture( aPicture, &dr );
	Unlock();
	
	SetGWorld( savePort, saveDevice );
}		

 
/***********************************************************************************************
*
*	SETPORTTOGW- returns the current port and device before setting it to this object
*
***********************************************************************************************/


void	ZGWorld::SetPortToGW( CGrafPtr* savedPort, GDHandle* savedDevice )
{
	// sets this object to be the current port, returning the previous setting. After a
	// sequence of calls, you should call SetGWorld() with the saved items. The copy functions
	// do this for you, so it is not required for those.
	
	GetGWorld( savedPort, savedDevice );
	SetGWorld( gw, NULL );
	
	// also set the gdRefCon to point to this, in case anyone (such as a search proc)
	// wants to find it
	
	GDHandle	gdH = GetGWorldDevice( gw );
	(*gdH)->gdRefCon = (long) this;
}


/***********************************************************************************************
*
*	SETPORTCOLOURS- sets the GWorld fore and back colour to those passed.
*
***********************************************************************************************/


void	ZGWorld::SetPortColours( RGBColor* aFore, RGBColor* aBack )
{
	// sets the GWorld's fore and back colours to those passed
	
	CGrafPtr	savePort;
	GDHandle	saveDevice;
	
	SetPortToGW( &savePort, &saveDevice );
	RGBForeColor( aFore );
	RGBBackColor( aBack );
	SetGWorld( savePort, saveDevice );
}



void	ZGWorld::SetPortColours()
{
	CGrafPtr	savePort;
	GDHandle	saveDevice;
	
	SetPortToGW( &savePort, &saveDevice );
	ForeColor( blackColor );
	BackColor( whiteColor );
	SetGWorld( savePort, saveDevice );
}


/***********************************************************************************************
*
*	GETPORTCOLOURS- gets the current GWorld fore and back colours.
*
***********************************************************************************************/

void	ZGWorld::GetPortColours( RGBColor* aFore, RGBColor* aBack )
{
	// returns the current port colours.
	
	CGrafPtr	savePort;
	GDHandle	saveDevice;
	
	SetPortToGW( &savePort, &saveDevice );
	GetForeColor( aFore );
	GetBackColor( aBack );
	SetGWorld( savePort, saveDevice );
}


/***********************************************************************************************
*
*	COPYICON- draw an icon into the GWorld.
*
***********************************************************************************************/

void	ZGWorld::CopyIcon( CIconHandle anIcon, Rect* dest, short align, short transform )
{
	CGrafPtr	savePort;
	GDHandle	saveDevice;
	
	FailNILParam( anIcon );
	
	Lock();
	SetPortToGW( &savePort, &saveDevice );

	PlotCIconHandle( dest, align, transform, anIcon );
	Unlock();
	SetGWorld( savePort, saveDevice );
}


/***********************************************************************************************
*
*	COPYICON- draw an icon into the GWorld. This searches for icons with the given ID
*
***********************************************************************************************/

void	ZGWorld::CopyIcon( short iconID, Rect* dest, short align, short transform )
{
	CIconHandle		anIcon;
	
	// is there a 'cicn' with this ID?
	
	anIcon = GetCIcon( iconID );
	
	if ( anIcon )
	{
		// yep- plot it
	
		CopyIcon( anIcon, dest, align, transform );
		DisposeCIcon( anIcon );
	}
	else
	{
		// no- is there an ICN# with this ID?
		
		CGrafPtr		savePort;
		GDHandle		saveDevice;
		OSErr			theErr;			
		
		Lock();
		SetPortToGW( &savePort, &saveDevice );
		
		theErr = PlotIconID( dest, align, transform, iconID );
		
		// if the error is not noErr, then the icon couldn't be found, so
		// try looking for a plain ICON resource
		
		if ( theErr )
		{
			Handle		iconH = GetIcon( iconID );
			
			if ( iconH )
			{
				PlotIcon( dest, iconH );
				ReleaseResource( iconH );
			}
		}
		
		Unlock();
	
		SetGWorld( savePort, saveDevice );
	}
}


/***********************************************************************************************
*
*	INVERT- invert the GWorld's image, with optional mask region
*
***********************************************************************************************/

void	ZGWorld::Invert( RgnHandle maskRgn )
{
	CGrafPtr		savePort;
	GDHandle		saveDevice;
	PixMapHandle	pp;
	Rect			r;
	
	// if depth is <16, invert colours in table rather than image
	
	if ( GetDepth() < 16 && GetDepth() > 1 )
	{
		CTabHandle	ct;
		short		i;
		
		ct = GetColours();
		
		for ( i = 0; i <= (*ct)->ctSize; i++ )
		{
			(*ct)->ctTable[i].rgb.red 	^= 0xFFFF;
			(*ct)->ctTable[i].rgb.green ^= 0xFFFF;
			(*ct)->ctTable[i].rgb.blue 	^= 0xFFFF;
		}
		
		CTabChanged( ct );
	}
	else
	{
		Lock();
		SetPortToGW( &savePort, &saveDevice );
			
		pp = GetPixMap();
		GetSize(&r);
		
		CopyBits((BitMap*) *pp, (BitMap*) *pp, &r, &r, notSrcCopy, maskRgn );
		
		Unlock();
		SetGWorld( savePort, saveDevice );
	}
}


/***********************************************************************************************
*
*	FLIPHORIZONTAL- flip the image horizontally, with optional mask region
*
***********************************************************************************************/

void	ZGWorld::FlipHorizontal( RgnHandle maskRgn )
{
	if ( maskRgn )
		FlipRegion( maskRgn, kFlipHorizontal );	
	
	Flip( kFlipHorizontal, maskRgn );		
}


/***********************************************************************************************
*
*	FLIPVERTICAL- flip the image vertically, with optional mask region
*
***********************************************************************************************/

void	ZGWorld::FlipVertical( RgnHandle maskRgn )
{
	if ( maskRgn )
		FlipRegion( maskRgn, kFlipVertical );	
	
	Flip( kFlipVertical, maskRgn );		
}


/***********************************************************************************************
*
*	FLIP- flip the entire image horizontally or vertically
*
***********************************************************************************************/

 void	ZGWorld::Flip( const short flipDir, RgnHandle mask )
{
	Rect		src, dest, fr;
	short		loops, i;
	
	ZGWorld*	temp;
	
	SetWatchCursor();
	FailNIL( temp = new ZGWorld( this ));
	
	// copy image back to this, one h or v strip at a time according to direction
	
	GetSize( &fr );
	src = dest = fr;
	
	if ( flipDir == kFlipHorizontal )
	{
		src.right = src.left + 1;
		dest.left = dest.right - 1;
		loops = fr.right - fr.left;	
	}
	else
	{
		src.bottom = src.top + 1;
		dest.top = dest.bottom - 1;
		loops = fr.bottom - fr.top;	
	}
	
	for( i = 0; i < loops; i++ )
	{
		CopyIn( temp, &src, &dest, srcCopy, mask );
		
		if ( flipDir == kFlipHorizontal )
		{
			OffsetRect( &src, 1, 0 );
			OffsetRect( &dest, -1, 0 );
		}
		else
		{
			OffsetRect( &src, 0, 1 );
			OffsetRect( &dest, 0, -1 );
		}
	}
	
	ForgetObject( temp );
}


/***********************************************************************************************
*
*	INSTALLSEARCHPROC- add a search procedure to this GWorld's device
*
***********************************************************************************************/

void	ZGWorld::InstallSearchProc( ColorSearchUPP aSearchProc )
{
	CGrafPtr	savePort;
	GDHandle	saveDevice;
	
	SetPortToGW( &savePort, &saveDevice );
	AddSearch( aSearchProc );
	SetGWorld( savePort, saveDevice );
}


/***********************************************************************************************
*
*	REMOVESEARCHPROC- add a search procedure to this GWorld's device
*
***********************************************************************************************/

void	ZGWorld::RemoveSearchProc( ColorSearchUPP aSearchProc )
{
	CGrafPtr	savePort;
	GDHandle	saveDevice;
	
	SetPortToGW( &savePort, &saveDevice );
	DelSearch( aSearchProc );
	SetGWorld( savePort, saveDevice );
}



/***********************************************************************************************
*	PROTECTED METHODS
***********************************************************************************************/




/***********************************************************************************************
*
*	MAKEMACGWORLD- makes a new GWorld with the characteristics passed
*
***********************************************************************************************/

void	ZGWorld::MakeMacGWorld( const Rect& aSize,
								const short aDepth,
								const CTabHandle aCTable,
								const Boolean makeTemp )
{
	// creates the GWorld. Initially, the pixels are unlocked and non-purgeable.
	
	
	GDHandle		gdH;
	GWorldFlags		gwFlags = 0;
	
	if ( makeTemp )
		gwFlags = useTempMem;
	
	FailOSErr( NewGWorld( &gw, aDepth, &aSize, aCTable, NULL, gwFlags ));
	pState = GetPixelsState( GetGWorldPixMap( gw ));
	
	gdH = GetGWorldDevice( gw );
	(*gdH)->gdRefCon = (long) this;
	
	// set initial clip and vis regions to the bounds
	
	RectRgn( gw->visRgn, &aSize );
	CopyRgn( gw->visRgn, gw->clipRgn );
}


/***********************************************************************************************
*
*	MAKEGWORLDFROMPICTURE- makes a GWorld from a picture handle.
*
***********************************************************************************************/

void	ZGWorld::MakeGWorldFromPicture( const PicHandle aPic, const Boolean makeTemp )
{
	// makes a GWorld that matches the picture's dimensions and draws the picture there
	
	char		pHState;
	Rect		aSize;
	short		aDepth;
	CTabHandle	aCTable = NULL;
	
	pHState = HGetState((Handle) aPic );
	HNoPurge((Handle) aPic );
	
	try
	{
		GetPicParams( aPic, &aSize, &aDepth, &aCTable );
		MakeMacGWorld( aSize, aDepth, aCTable, makeTemp );
		
		// draw the picture in the GWorld
		
		Clear();
		CopyPicture( aPic, &aSize );	
	}
	catch( OSErr err )
	{
		if ( aCTable )
			DisposeCTable( aCTable );
		
		HSetState((Handle) aPic, pHState );
		
		throw err;
	}

	if ( aCTable )
		DisposeCTable( aCTable );
	
	HSetState((Handle) aPic, pHState );
}


/***********************************************************************************************
*
*	SETMACGWORLD- makes this object manage an"external" GWorld
*
***********************************************************************************************/

void	ZGWorld::SetMacGWorld( GWorldPtr aGW )
{
	gw = aGW;
}


/***********************************************************************************************
*
*	WRITETOSTREAM- store the GWorld in the stream
*
***********************************************************************************************/

void	ZGWorld::WriteToStream( ZStream* aStream )
{
	// in order to stream a GWorld, we need to save sufficient parameters from the pixMap
	// to specify the image, plus the image data itself. This amounts to quite a lot!
	
#if _MACZOOP_STREAMS
	PixMapHandle	pp;
	
	pp = GetPixMap( TRUE );
	
	try
	{
		aStream->WriteRect( &(*pp)->bounds );
		aStream->WriteShort((*pp)->pixelSize );
		aStream->WriteHandle((Handle)(*pp)->pmTable );
		aStream->WriteChar( inTempMem );
		
		// additional pixmap parameters

		aStream->WriteShort((*pp)->rowBytes );
		aStream->WriteShort((*pp)->pixelType );
		aStream->WriteShort((*pp)->cmpCount );
		aStream->WriteShort((*pp)->cmpSize );
		aStream->WriteLong((long)(*pp)->hRes );
		aStream->WriteLong((long)(*pp)->vRes );

		// stream out the actual pixels
		
		aStream->WriteData((*pp)->baseAddr, GetImageMemSize());
	}
	catch( OSErr err )
	{
		Unlock();
		throw err;
	}
	
	Unlock();
#endif
}


/***********************************************************************************************
*
*	READFROMSTREAM- rebuild the GWorld from the stream
*
***********************************************************************************************/

void	ZGWorld::ReadFromStream( ZStream* aStream )
{
#if _MACZOOP_STREAMS
	Rect		br;
	short		depth, rb, pt, cc, cs;
	CTabHandle	ct;
	long		ctm;
	Fixed		hr, vr;
	
	aStream->ReadRect( &br );
	aStream->ReadShort( &depth );
	aStream->ReadHandle((Handle*) &ct );
	aStream->ReadChar((char*) &inTempMem );
	
	MakeMacGWorld( br, depth, ct, inTempMem );
	
	if ( ct )
		DisposeHandle((Handle) ct );
	
	// now check the parameters and stream in the pixel data
	
	PixMapHandle	pp = GetPixMap( TRUE );
	
	aStream->ReadShort( &rb );
	aStream->ReadShort( &pt );
	aStream->ReadShort( &cc );
	aStream->ReadShort( &cs );
	aStream->ReadLong((long*) &hr );
	aStream->ReadLong((long*) &vr );
	
	(*pp)->hRes = hr;
	(*pp)->vRes = vr;
	
	// the crucial thing is that the pixel data saved and the amount of space available
	// in the GWorld to store them is the same. We can be sure about the bounds rect since
	// we explicitly passed it, so we only need to compare rowBytes.
	
	if ( rb == (*pp)->rowBytes )
		aStream->ReadData((*pp)->baseAddr, &ctm );
	else
	{
		// if the rowBytes didn't match, we're up a pole. However, in order to be as graceful
		// as possible, we skip the image data in the stream in order to keep things going
		// as best we can. OK, the image will be blank, but nothing very bad will happen. In
		// normal use this should never occur.
		
		aStream->ReadLong( &ctm );
		aStream->Skip( ctm );
	}
	
	Unlock();
#endif	
}


/***********************************************************************************************
*
*	GETPICPARAMS- extracts pixmap info from a picture. This is not a method.
*
***********************************************************************************************/

typedef struct
{
	Boolean		iValid;			// TRUE if bits proc was called at least once
	short		iDepth;			// depth of the bitmap or pixmap
	CTabHandle	iColours;		// the colour table of the pixmap (a copy)
}
PPicInfo;

// global needed to communicate between method and bottleneck proc

static	PPicInfo	gPPInfo;

// bottleneck proc to extract the info

static  pascal void	PPGetBits(BitMap *srcBits, Rect *srcRect, Rect *dstRect, short mode, RgnHandle maskRgn);
static  pascal void	PPGetBits(BitMap *srcBits, Rect *srcRect, Rect *dstRect, short mode, RgnHandle maskRgn)
{
	// first let's see if this is a colour pixmap or not
	
	PixMapPtr	pp;
	Boolean		isColour;
	
	pp = (PixMapPtr) srcBits;
	
	// if the top bit of rowBytes is set, this is a colour port
	
	isColour = ((pp->rowBytes & 0x8000) != 0);
	
	if (isColour)
	{
		// Set the depth of the global record if this pixmap's depth
		// is greater than any seen so far
		
		if (pp->pixelSize > gPPInfo.iDepth)
			gPPInfo.iDepth = pp->pixelSize;
			
		// if the record has not been previously set, set up the colour table
		
		if (!gPPInfo.iValid && pp->pmTable)
		{
			gPPInfo.iColours = pp->pmTable;
		
			// make a copy of the table
		
			HandToHand((Handle*) &gPPInfo.iColours);
			gPPInfo.iValid = TRUE;	// don't set the colours again
		}
	}
	
	// note we do not call StdBits, since we are only interested in the information, and couldn't
	// really care less about actually drawing it. This also results in much faster operation.
}



void	GetPicParams(PicHandle aPic, Rect* aSize, short* aDepth, CTabHandle* aCTable)
{
	// return the parameters of the picture. This can be problematic- the picture encapsulates its data in
	// such a way that quickly getting depth and colours is not straightforward. Also, a given picture may
	// not contain a pixmap and so such terms are meaningless. However, we do our best. This method works by
	// installing a bottleneck procedure into a temporary port and looking for CopyBits opcodes. It then
	// extracts the parameters passed to that call (depth and colours). If no colour pixmap is encountered,
	// the depth returned is always 1, and the colours NULL. The size returned is always valid.
	
	CGrafPort	tempPort;
	CGrafPtr	savePort, tPort;
	CQDProcs	tProcs;
	Rect		r;
	
	// the size is the one piece of info we can extract directly
	
	r = (*aPic)->picFrame;
	OffsetRect(&r, -r.left, -r.top);
	*aSize = r;
	
	// open a port
	
	GetPort((GrafPtr*) &savePort);
	OpenCPort(&tempPort);
	tPort = &tempPort;
	SetPort((GrafPtr) tPort);
	
	// make the clip and vis regions empty so that the call to draw picture doesn't draw anything on
	// the screen by mistake
	
	SetEmptyRgn(tempPort.visRgn);
	SetEmptyRgn(tempPort.clipRgn);
	
	// install a bottleneck procedure for intercepting CopyBits
	
	SetStdCProcs(&tProcs);
	
	// make the StdBits pointer point to our function
	
	tProcs.bitsProc = NewQDBitsProc(PPGetBits);
	gPPInfo.iValid = FALSE;
	gPPInfo.iDepth = 1;
	gPPInfo.iColours = NULL;
	
	// put it in the port
	
	tPort->grafProcs = &tProcs;
	
	// now draw the picture in the port, which should result in our bottleneck getting called at least once,
	// provided the picture does contain a pixmap or bitmap.
	
	DrawPicture(aPic, &r);
	
	// now we should have the required info in the global gPPInfo record. First however, lets clean up.
	
	tPort->grafProcs = NULL;
	DisposeRoutineDescriptor(tProcs.bitsProc);
	CloseCPort(tPort);
	SetPort((GrafPtr) savePort);
	
	// now simply return the data
	
	*aDepth = gPPInfo.iDepth;
	*aCTable = gPPInfo.iColours;
}





void	Scale2Rects(Rect *theRect,Rect *refRect)
{
	// this function resizes <theRect> so that it will fit into <refRect>, but preserving its
	// original aspect ratio. This is useful for preventing unwanted stretching of pictures
	// both for creation and display of preview images. Note this also centres the rect within
	// the refRect for the shorter dimension. This is the more intuitively correct behaviour

	Fixed 	aspectRatio;
	Rect	destFrame;
	short	refWidth,refHeight;
	
	destFrame.top = destFrame.left = 0;
	OffsetRect(theRect,-theRect->left,-theRect->top);
	
	refWidth = refRect->right - refRect->left;
	refHeight = refRect->bottom - refRect->top;
	
	if (theRect->right < theRect->bottom)
	{
		// the rect is taller than it is wide, so we centre horizontally
		
		aspectRatio = FixRatio(theRect->right,theRect->bottom);
		destFrame.right = FixRound(FixMul(aspectRatio,FixRatio(refHeight,1)));
		destFrame.bottom = refHeight;
		
		OffsetRect(&destFrame,refRect->left + ((refWidth - destFrame.right) / 2),refRect->top);
	}
	else
	{
		// the rect is wider than it is tall, so we centre vertically 
		
		aspectRatio = FixRatio(theRect->bottom,theRect->right);
		destFrame.bottom = FixRound(FixMul(aspectRatio,FixRatio(refWidth,1)));
		destFrame.right = refWidth;
		
		OffsetRect(&destFrame,refRect->left,refRect->top + ((refHeight - destFrame.bottom) / 2));
	}
	*theRect = destFrame;
}





void	FlipRegion( RgnHandle aRgn, const short flipDir )
{
	// flip a region horizontally or vertically. Converts to a bitmap,
	// flips it and converts back.
	
	Rect		r;
	ZGWorld*	rMap;
	CGrafPtr	savePort;
	GDHandle	saveDev;
	
	FailNILParam( aRgn );
	
	r = (*aRgn)->rgnBBox;
	FailOSErr( EmptyRect( &r )? paramErr: noErr );
	
	// if the region is rectangular, don't bother
	
	if ( GetHandleSize((Handle) aRgn ) != 10 )
	{	
		// make 1-bit world in temp memory
		
		FailNIL( rMap = new ZGWorld( r, 1, NULL, TRUE ));
		
		try
		{
			rMap->Lock();
			rMap->SetPortToGW( &savePort, &saveDev );
			
			PaintRgn( aRgn );
			
			// flip it
			
			if ( flipDir == kFlipHorizontal )
				rMap->FlipHorizontal();
			else
				rMap->FlipVertical();
				
			// convert back to region
			PixMapHandle	pp = rMap->GetPixMap();
			FailOSErr( BitMapToRegion( aRgn, (BitMap*) *pp ));
			
			rMap->Unlock();
			SetGWorld( savePort, saveDev );
			ForgetObject( rMap );
		}
		catch( OSErr err )
		{
			rMap->Unlock();
			SetGWorld( savePort, saveDev );
			ForgetObject( rMap );
			
			throw err;
		}
	}
}
