/*************************************************************************************************
*
*
*			MacZoop - "the framework for the rest of us"		 
*
*
*
*			ZGWorldWindow.cpp	-- a window with an offscreen GWorld. This can open PICT
*									images into an offscreen port and correctly uses the
*									palette manager, etc
*
*
*
*
*
*			 1996, Graham Cox
*
*
*
*
*************************************************************************************************/



#include	"ZGWorldWindow.h"
#include	"ZGWorld.h"
#include	"ZUndoIPTask.h"
#include	"MacZoop.h"

#include	<Palettes.h>
#include	<PictUtils.h>
#include	<drag.h>


CLASSCONSTRUCTOR( ZGWorldWindow );

/*-------------------------------***  CONSTRUCTOR  ***----------------------------------*/

ZGWorldWindow::ZGWorldWindow( ZCommander* aBoss, const short windID )
	: ZPictWindow( aBoss, windID )
{
	classID = CLASS_ZGWorldWindow;
	
	itsOffscreen = NULL;
	printable = TRUE;
	scale = 100;
	fatBitsGrid = FALSE;
}


ZGWorldWindow::ZGWorldWindow()
	: ZPictWindow()
{
	classID = CLASS_ZGWorldWindow;
	
	itsOffscreen = NULL;
	printable = TRUE;
	scale = 100;
	fatBitsGrid = FALSE;
}

/*--------------------------------***  DESTRUCTOR  ***---------------------------------*/

ZGWorldWindow::~ZGWorldWindow()
{
	if (itsOffscreen)
		ForgetObject(itsOffscreen);
}


/*-------------------------------***  DRAWCONTENT  ***----------------------------------*/
/*	

render the GWorld into the window

----------------------------------------------------------------------------------------*/

void	ZGWorldWindow::DrawContent()
{
	Rect	r, ir, vr, sr;
	short	ff;
	
	GetBounds( &r );
	
	vr = (*qd.thePort->visRgn)->rgnBBox;
	
	if (itsOffscreen)
	{
		SectRect(&r, &vr, &r );
		
		// to prevent visual artifacts, the window rect needs to be made up to
		// enclose an exact whole number of pixels at the zoomed size. This can
		// occur at zoomed up scales.
		
		if ( isPrinting )
		{
			itsOffscreen->GetSize( &sr );
			r = sr;
			OffsetRect( &r, -r.left, -r.top );
		}
		else
		{
			if ( scale > 100 )
			{
				ff = scale / 100;
				
				r.left -= r.left % ff;
				r.top -= r.top % ff;
				r.right += ff - ( r.right % ff );
				r.bottom += ff - ( r.bottom % ff );
			}
			
			CalcSourceRect( &vr );
			GetContentRect( &ir );
			
			sr = r;
			MapRect( &sr, &ir, &vr );
		}
		ForeColor( blackColor );
		BackColor( whiteColor );

		itsOffscreen->CopyOut( &qd.thePort->portBits, &sr, &r, srcCopy + ditherCopy );
			
		if ( fatBitsGrid && ( scale > 400 ) && ! isPrinting)
			CalcFatBitsGridMask( r );
			
		// if window bigger than image, erase the remainder
		
		RectRgn( gUtilRgn, &r );
		DiffRgn( qd.thePort->visRgn, gUtilRgn, gUtilRgn );
		EraseRgn( gUtilRgn );
		
	}
	else
		ZPictWindow::DrawContent();
}


/*-----------------------------***  HANDLECOMMAND  ***----------------------------------*/
/*	

handle the "Invert" command
----------------------------------------------------------------------------------------*/

void	ZGWorldWindow::HandleCommand( const long theCommand )
{
	switch ( theCommand )
	{
		case kCmdInvert:
			if ( itsOffscreen )
			{
				ZUndoIPTask*	aTask;
				Str63			taskStr;
				
				// make a task to actually handle the command. This allows it
				// to be undoable.
				
				GetIndString( taskStr, kIPTasksStrResID, undoInvert );
				FailNIL( aTask = new ZUndoIPTask( taskStr, this, undoInvert ));
				
				// do the task initially
				
				aTask->Do();
				
				// send the task up to the app so that it becomes the current
				// undo menu item. Also marks the document as dirty.
				
				SetTask( aTask );
			}
			break;
		case kCmdZoomIn:
			ZoomToCentre( FALSE );
			break;
		case kCmdZoomOut:
			ZoomToCentre( TRUE );
			break;
		default:
			ZPictWindow::HandleCommand( theCommand );
			break;
	}
}



/*-------------------------------***  UPDATEMENUS  ***----------------------------------*/
/*	

enable the "Invert" command

----------------------------------------------------------------------------------------*/

void	ZGWorldWindow::UpdateMenus()
{
	if (itsOffscreen)
	{
		gMenuBar->EnableCommand( kCmdInvert );
		gMenuBar->EnableCommand( kCmdCopy );
		gMenuBar->EnableCommand( kCmdZoomIn );
		gMenuBar->EnableCommand( kCmdZoomOut );
	}
	
	ZPictWindow::UpdateMenus();

}

/*---------------------------------***  DOCOPY  ***-------------------------------------*/
/*	
put the image on the clipboard as a PICT
----------------------------------------------------------------------------------------*/

void	ZGWorldWindow::DoCopy()
{
	// remove any existing picture
	
	FailNIL(itsOffscreen);
	if (thePicture)
		KillPicture(thePicture);
	
	// make a new one from the image
	
	FailNIL(thePicture = itsOffscreen->MakePicture());
	
	try
	{
		// call the inherited method from ZPictWindow to do the work
		
		ZPictWindow::DoCopy();
	}
	catch(OSErr err)
	{
		KillPicture(thePicture);
		thePicture = NULL;
		throw err;
	}
	
	// clean up
	
	KillPicture(thePicture);
	thePicture = NULL;
}

/*--------------------------------***  DOPASTE  ***-------------------------------------*/
/*	
paste any PICT on the clipboard as a new GWorld
----------------------------------------------------------------------------------------*/

void	ZGWorldWindow::DoPaste()
{
	// make an empty handle for putting the picture in
	
	thePicture = (PicHandle) NewHandle(0);
	
	// call ZPictWindow to copy the clipboard contents
	
	ZPictWindow::DoPaste();
	
	// convert to a GWorld
	
	if (thePicture)
	{
		try
		{
			MakeGWorld();
		
			KillPicture(thePicture);
			thePicture = NULL;
			
			Focus();
			Draw();
		}
		catch(OSErr err)
		{
			// if MakeGWorld failed, it will have thrown an exception. In
			// this case we do NOT propagate it, but instead let the subclass
			// handle the picture directly
		}
	}
}



/*-----------------------------***  ACCEPTSFLAVOUR  ***---------------------------------*/
/*	

tell drag manager what we can receive- PICT clippings and files of type PICT

----------------------------------------------------------------------------------------*/

Boolean	ZGWorldWindow::AcceptsFlavour( const OSType aFlavour )
{
	return (aFlavour == 'PICT' ||
			aFlavour == 'hfs ');
}



/*-----------------------------------***  DROP  ***-------------------------------------*/
/*	

handle drag manager drops of picture clippings, etc

----------------------------------------------------------------------------------------*/

void	ZGWorldWindow::Drop( const OSType flavour, const Ptr data, const long dataSize )
{
	// <data> points to a pict- make a real picture handle by copying it
	
	if (flavour == 'PICT')
	{
		FailOSErr(PtrToHand(data, (Handle*) &thePicture, dataSize));
		
		// make a new GWorld from the picture
		
		MakeGWorld();
		
		short	sm = ( 8 * scale ) / 100;
		SetScrollAmount( sm, sm );
		
		// get rid of the picture- no longer needed
		
		KillPicture(thePicture);
		thePicture = NULL;
	}
	else
	{
		if (flavour == 'hfs ')
		{
			HFSFlavor*	fSpec;
			
			fSpec = (HFSFlavor*) data;
			
			if (fSpec->fileType == 'PICT')
			{
				// it's a PICT file, so try and open it
		
				SetFile( fSpec->fileSpec );
				OpenFile( fSpec->fileType );
			}
			else
				FailOSErr(paramErr);	// not a file we can open
		}
	}
	// update everything
	
	Focus();
	Draw();
}


/*---------------------------------***  OPENFILE  ***------------ ----------------------*/
/*	

open a PICT file and make a GWorld out of it

----------------------------------------------------------------------------------------*/

void	ZGWorldWindow::OpenFile( const OSType fType, Boolean isStationery )
{
	// call our ancestor to do the hard work of opening the file, etc
	
	ZPictWindow::OpenFile( fType, isStationery );
	
	// if OK, then make a GWorld from the picture
	
	if ( thePicture )
	{
		try
		{
			MakeGWorld();
		
			KillPicture( thePicture );
			thePicture = NULL;
		}
		catch( OSErr err )
		{
			// if MakeGWorld failed, it will have thrown an exception. In
			// this case we do NOT propagate it, but instead let the subclass
			// handle the picture directly
		}
	}
}


/*---------------------------------***  SAVEFILE  ***------------ ----------------------*/
/*	

save a PICT file from the offscreen GWorld

----------------------------------------------------------------------------------------*/

void	ZGWorldWindow::SaveFile()
{
	// if the offscreen exists, kill any existing picture and make a new one. Otherwise
	// use whatever picture is there.
	
	if ( itsOffscreen )
	{
		if ( thePicture )
			KillPicture( thePicture );

		FailNIL( thePicture = itsOffscreen->MakePicture());
	}
	// call our ancestor to save the picture to disk
	
	try
	{
		ZPictWindow::SaveFile();
	}
	catch( OSErr err )
	{
		if ( itsOffscreen )
		{
			KillPicture( thePicture );
			thePicture = NULL;
		}
		
		throw err;
	}
	
	// get rid of the picture if there's an offscreen image
	// - we don't need it any longer
	
	if ( itsOffscreen )
	{
		KillPicture( thePicture );
		thePicture = NULL;
	}
}


/*----------------------------------***  GETNAME  ***-----------------------------------*/
/*	
this window stores the name in the filespec, since the window title has extra info.
----------------------------------------------------------------------------------------*/

void	ZGWorldWindow::GetName( Str255 name )
{
	CopyPString( macFile.name, name );
}


/*---------------------------------***  SETTITLE  ***-----------------------------------*/
/*	
the window title is the file name + the zoom factor as a percentage.
----------------------------------------------------------------------------------------*/

void	ZGWorldWindow::SetTitle( Str255 name )
{
	Str63	wtName;
	Str15	zfStr;
	
	CopyPString( name, macFile.name );
	CopyPString( name, wtName );
	
	NumToString( scale, zfStr );
	ConcatPStrings( wtName, "\p @");
	ConcatPStrings( wtName, zfStr );
	ConcatPStrings( wtName, "\p%" );
	
	SetWTitle( macWindow, wtName );
}
	



/*--------------------------------***  MAKEGWORLD  ***----------------------------------*/
/*	
construct the GWorld from the picture in <thePicture>
----------------------------------------------------------------------------------------*/

void	ZGWorldWindow::MakeGWorld()
{
	Rect	aSize;
	
	FailNIL(thePicture);
	
	if (itsOffscreen)
		ForgetObject(itsOffscreen);
			
	FailNIL(itsOffscreen = new ZGWorld(thePicture));
	
	CalcBoundsRect( &aSize );
	SetBounds( aSize );
	
	MakePaletteForWindow();
}


/*--------------------------------***  SETGWORLD  ***-----------------------------------*/
/*	
set the window to manage a new GWorld
----------------------------------------------------------------------------------------*/

void	ZGWorldWindow::SetGWorld( ZGWorld* aGW )
{
	itsOffscreen = aGW;
	SendMessage( kMsgGWorldChanged, (void*) itsOffscreen );
}


/*---------------------------***  MAKEPALETTEFORWINDOW  ***-----------------------------*/
/*	
install a palette in the window that best represents the image on a CLUT device.
----------------------------------------------------------------------------------------*/

void	ZGWorldWindow::MakePaletteForWindow()
{
	short			gwDepth;
	CTabHandle		aCTab;
	PaletteHandle	aPalette;
	
	FailNIL(itsOffscreen);
	
	gwDepth = itsOffscreen->GetDepth();
	
	if (gwDepth > 8)
	{
		// for a direct image, there is no colour table immediately available, so we
		// need to make one that has the best colours. We can do this by calling the
		// Picture Utilities package to give us a reasonable table.
	
		PictInfo		pi;
		PixMapHandle	pp;
		
		pp = itsOffscreen->GetPixMap(TRUE);		// lock the GWorld as well
		
		try
		{
			FailOSErr(GetPixMapInfo(pp, &pi, returnPalette + suppressBlackAndWhite, 254, systemMethod, 0));
		}
		catch(OSErr err)
		{
			itsOffscreen->Unlock();
			
			throw err;
		}
		itsOffscreen->Unlock();
		aPalette = pi.thePalette;
	}
	else
	{
		// the image is already an indexed image, so use its colour table
	
		aCTab = itsOffscreen->GetColours();
		
		// make a palette from the CTable
	
		aPalette = NewPalette(256, aCTab, pmTolerant, 0);	
	}
	
	// install the palette
	
	SetPalette(macWindow, aPalette, TRUE);
}


void	ZGWorldWindow::SetScale( const short aScale )
{
	short	sc = MAX( 25, MIN( aScale, 3200 ));
	
	if ( sc != scale )
	{
		//scale can be from 25% to 3,200%
		
		scale = sc;
	
		Rect	br;
		
		CalcBoundsRect( &br );
		SetBounds( br );

		// scroll amounts should be factored to give consistent scroll rates
		
		short	sm = ( 8 * scale ) / 100;
		SetScrollAmount( sm, sm );
		
		SetTitle( macFile.name );
	}
}


void	ZGWorldWindow::ZoomToPoint( const Point clickPt, Boolean isZoomOut )
{
	// zoom in or out so that the passed point is centred in the window is possible. This is
	// the basis of a "magnifying" tool for doing zooming.
	
	Rect	cont, obr, br;
	short	scrollH, scrollV, newScale;
	Point	cPt;
	
	Focus();
	GetContentRect( &cont );
	GetBounds( &obr );
	
	newScale = isZoomOut? ( scale >> 1 ) : ( scale << 1 );
	
	// if drag manager is available, perform a zoom effect
	
	newScale = MAX( 25, MIN( newScale, 3200 ));
	
	if ( newScale != scale )
	{
		if ( gMacInfo.hasDragManager )
		{
			Rect	zSrc, zDest;
			short	dh, dv;
			
			GetContentRect( &zDest );
			Pt2Rect( clickPt, clickPt, &zSrc );
			InsetRect( &zSrc, -2, -2 );
			GetPosition( &dh, &dv );
			OffsetRect( &zSrc, -dh, -dv );
			
			LocalToGlobal( &topLeft( zSrc ));
			LocalToGlobal( &botRight( zSrc ));
			LocalToGlobal( &topLeft( zDest ));
			LocalToGlobal( &botRight( zDest ));
			
			if ( isZoomOut )
				FailOSErr( ZoomRects( &zDest, &zSrc, 10, kZoomDecelerate ));
			else
				FailOSErr( ZoomRects( &zSrc, &zDest, 10, kZoomAccelerate ));
		}
		
		SetScale( newScale );
		GetBounds( &br );
		
		cPt = clickPt;
		MapPt( &cPt, &obr, &br );
		
		scrollH = cPt.h - (( cont.right - cont.left ) / 2 );
		scrollV = cPt.v - (( cont.bottom - cont.top ) / 2 );
		
		SetControlValue( theHBar, scrollH );
		SetControlValue( theVBar, scrollV );
		SetOriginToScroll();
		DrawContent();
	}
	else
		SysBeep( 1 );	// hit the buffers!
}


void	ZGWorldWindow::ZoomToCentre( Boolean isZoomOut )
{
	Rect	br;
	Point	clickPt;
	
	GetBounds( &br );
	
	clickPt.h = ( br.left + br.right ) / 2;
	clickPt.v = ( br.top + br.bottom ) / 2;
	
	ZoomToPoint( clickPt, isZoomOut );
}


void	ZGWorldWindow::CalcSourceRect( Rect* aSrcRect )
{
	// computes the rect that is the corresponding area in the offscreen world that is
	// displayed in the aperture of the window.
	
	Rect	ap, br, r;
	
	if ( itsOffscreen )
	{
		CalcBoundsRect( &br );
		GetContentRect( &ap );
		
		itsOffscreen->GetSize( &r );
		MapRect( &ap, &br, &r );
	}
	else
		SetRect( &ap, 0, 0, 0, 0 );
		
	*aSrcRect = ap;
}



void	ZGWorldWindow::CalcBoundsRect( Rect* aBoundsRect )
{
	Rect	r;
	
	if ( itsOffscreen )
	{
		itsOffscreen->GetSize( &r );

		r.right = ((long) r.right * (long) scale ) / 100;
		r.bottom = ((long) r.bottom * (long) scale ) / 100;
	}
	else
		GetBounds( &r );
		
	*aBoundsRect = r;
}



void	ZGWorldWindow::CalcFatBitsGridMask( const Rect& destRect )
{
	short		x, y, ff;
	
	ff = scale / 100;	// display pixels per pixel of projected image
	
	PenNormal();
	ForeColor( whiteColor );
	
	for( x = destRect.left; x < destRect.right; x += ff )
	{
		MoveTo( x, destRect.top );
		LineTo( x, destRect.bottom );
	}
	
	for( y = destRect.top; y < destRect.bottom; y += ff )
	{
		MoveTo( destRect.left, y );
		LineTo( destRect.right, y );
	}
	
	ForeColor( blackColor );
}


void	ZGWorldWindow::WriteToStream( ZStream* aStream )
{
#if _MACZOOP_STREAMS
	ZScroller::WriteToStream( aStream );
	
	aStream->WriteShort( scale );
	aStream->WriteChar( fatBitsGrid );
	aStream->WriteObject( itsOffscreen );
	
	if ( itsOffscreen == NULL )
		aStream->WriteHandle((Handle) thePicture );
#endif
}


void	ZGWorldWindow::ReadFromStream( ZStream* aStream )
{
#if _MACZOOP_STREAMS
	ZScroller::ReadFromStream( aStream );
	
	aStream->ReadShort( &scale );
	aStream->ReadChar((char*) &fatBitsGrid );
	itsOffscreen = (ZGWorld*) aStream->ReadObject();
	
	if ( itsOffscreen == NULL )
	{
		aStream->ReadHandle((Handle*) &thePicture );
	
		// if OK, then make a GWorld from the picture
		
		if ( thePicture )
		{
			try
			{
				MakeGWorld();
			
				KillPicture( thePicture );
				thePicture = NULL;
			}
			catch( OSErr err )
			{
				// if MakeGWorld failed, it will have thrown an exception. In
				// this case we do NOT propagate it, but instead let the subclass
				// handle the picture directly
			}
		}
	}
#endif
}
