/*************************************************************************************************
*
*
*			MacZoop - "the framework for the rest of us"		 
*
*
*
*			ZStdPrefsDialog.cpp	-- a standard multi-panel preferences dialog.
*
*
*			 1997, Graham Cox
*
*
*
*************************************************************************************************/

#include	"ZStdPrefsDialog.h"
#include	"ZPrefsFile.h"
#include	"ZExtraDialogItems.h"
#include	"MacZoop.h"
#include	"ZObjectArray.cpp"



extern ZPrefsFile*		gPrefsFile;

static ZArray*			gPrefMapRegistry = NULL;
	


ZStdPrefsDialog::ZStdPrefsDialog( ZCommander* aBoss, const short dialogID )
	: ZAdvancedDialog( aBoss, dialogID )
{
	maps = NULL;
	curPanel = 0;
}


ZStdPrefsDialog::~ZStdPrefsDialog()
{
	if ( maps )
	{
		maps->DisposeAll();
		ForgetObject( maps );
	}
}


	
void			ZStdPrefsDialog::SetUp()
{
	SetWatchCursor();
	SetDialogBaseFont( kFontIDGeneva, 9 );
	
	ZAdvancedDialog::SetUp();
	ParseDescriptorResource( windID );
	
	// initially the "Save" button is disabled- subsequent changes enable it
	
	DisableItem( kStdPrefSaveButton );
	
	// start off with the panel we had last time, if it still exists. Otherwise
	// switch to panel 1.
	
	short initialPanel = 1;
	
	if ( gPrefsFile )
	{
		PrefDPrefHdl	ppH = (PrefDPrefHdl) gPrefsFile->ReadResource( kPrefDialogPrefResType, 128 );
	
		if ( ppH )
		{
			initialPanel = (*ppH)->openPanelIndex;
			
			ReleaseResource((Handle) ppH );
		}
	}
		
	SwitchPanels( initialPanel );
	
	if ( DataToSave())
		EnableItem( kStdPrefSaveButton );
}


void			ZStdPrefsDialog::ClickItem( const short item )
{
	ZAdvancedDialog::ClickItem( item );
	
	switch ( item )
	{
		case kStdPrefSwitchIconList:
			break;
			
		default:
			if ( item > baseItems )
			{
				ZPrefMapper*	aMap = GetCurrentMapper();
				
				if ( aMap )
					aMap->PanelClick( item - baseItems );
					
				if ( DataToSave())
					EnableItem( kStdPrefSaveButton );
				else
					DisableItem( kStdPrefSaveButton );
			}
			break;
	}
}


Boolean		ZStdPrefsDialog::CloseDialog()
{
	// user is closing the dialog and accepting changes, so save all
	// the changed data back to the prefs file
	
	// we'd also like to save our own pref info so the right panel can be
	// restored next time.
	
	if ( gPrefsFile )
	{
		PrefDPrefHdl	pH = ( PrefDPrefHdl ) NewHandleClear( sizeof( PrefDPrefRec ));
		
		if ( pH )
		{
			ZPrefMapper*	aMap = GetCurrentMapper();
			
			(*pH)->version = 1;
			(*pH)->openPanelIndex = curPanel;
			(*pH)->openPanelClass = aMap->GetMapClass();
		
			gPrefsFile->WriteResource((Handle) pH, kPrefDialogPrefResType, 128 );
		}
	}		

	// save client data

	SwitchPanels( 0 );
	
	if ( DataToSave())	
		SaveChangedPrefData();
	
	return ZAdvancedDialog::CloseDialog();
}


void		ZStdPrefsDialog::ReceiveMessage( ZComrade* aSender, long msg, void* msgData )
{
	if ( msg == newListItemSelected )
		ProcessClickInIconList();	
}


void		ZStdPrefsDialog::ParseDescriptorResource( const short resID )
{
	// this does the majority of the setup. It reads the 'PINF' resource with the given ID and
	// tries to build a mapper for each class called for. If successful it sets up the mapper
	// and gets the pref data and builds the icon list.

	PrefDialogHdl		pdH;

	pdH = ( PrefDialogHdl ) GetResource( kPrefDialogDescResType, resID );
	
	FailNILErr( pdH, resNotFound );
	
	HNoPurge((Handle) pdH );
	
	// build the mappers list
	
	FailNIL( maps = new ZPrefMapperList());
	
	// make the required mappers
	
	short			i, iMax, pID, ditlID;
	OSType			mapClassID, pType;
	ZPrefMapper*	aMap;
	Handle			pData;
	IconInfo		info;
	
	iMax = (*pdH)->numPanels;
	
	ZIconListBox*	ilb = dynamic_cast<ZIconListBox*>( GetObjectItem( kStdPrefSwitchIconList ));	
	
	FailNILParam( ilb );
	
	ilb->SetCanBeHandler( FALSE );
		
	for( i = 0; i < iMax; i++ )
	{
		mapClassID = (*pdH)->pp[i].prefMapperClassID;
		
		// fill in the icon info record:
		
		info.iType = IconColour;	// currently only 'cicn' icons supported here
		info.iconID = (*pdH)->pp[i].iconID;
		info.theIcon = (Handle) GetCIcon( info.iconID );
		info.userData = (long) this;
		CopyPString((*pdH)->pp[i].prefPanelName, info.title );
		
		// if no icon could be loaded, skip this panel since without a way to select it,
		// there's no point having the map at all. In other words, the icon is required!
		
		if ( info.theIcon )
		{
			aMap = MakeMapper( mapClassID );
		
			// if mapper class unregistered, or otherwise NULL, just skip it.
			
			if ( aMap )
			{
				maps->AppendItem( aMap );
				
				// get the pref resource info from the descriptor record
			
				pType	= (*pdH)->pp[i].prefResType;
				pID		= (*pdH)->pp[i].prefResID;
				ditlID	= (*pdH)->pp[i].ditlID;
				pData	= NULL;
				
				// try to read the data from the prefs file. We always detach the resource since
				// the mappers store their state changes here. If the user cancels, we need to make
				// sure that the newer state is discarded.
				
				try
				{
					// try to get the resource. It's OK for there to be no prefs file- we might only
					// want to create one if the user changes from the default settings.
					
					if ( gPrefsFile )
						pData = gPrefsFile->ReadResource( pType, pID, TRUE );
				}
				catch( OSErr err )
				{
					pData = NULL;
				}

				if ( aMap )
					aMap->InitMapper( this, baseItems, pData, mapClassID, pType, pID, ditlID );
					
				if ( pData == NULL )
					pData = aMap->MakeDefaultPrefData();
					
				// set up the icon selector for this mapper
				
				if ( ilb )
					ilb->AppendIcon( &info );	
			}
		}
	}
	
	ReleaseResource((Handle) pdH );
}


ZPrefMapper*	ZStdPrefsDialog::MakeMapper( OSType classID )
{
	ZPrefMapper*	aMap = NULL;
	
	switch ( classID )
	{
		case kStdMapperClassID:
			aMap = new ZPrefMapper();
			break;
			
		default:
			// default case looks up mapper in registry	
			
			if ( gPrefMapRegistry )
			{
				long		i;
				MapInstRec	mi;
				
				for( i = 1; i <= gPrefMapRegistry->CountItems(); i++ )
				{
					gPrefMapRegistry->GetArrayItem( &mi, i );
					
					if ( mi.mapClassID == classID )
					{
						aMap = dynamic_cast<ZPrefMapper*>((*mi.inst)());
						break;
					}
				}
			}
			break;
	}
	
	return aMap;
}


void			ZStdPrefsDialog::SwitchPanels( const short panelID )
{
	ZPrefMapper*	aMap;
	ZIconListBox*	ilb;
	
	if ( panelID != curPanel )
	{
		aMap = GetCurrentMapper();
		
		if ( aMap )
			aMap->PanelTearDown();
			
		curPanel = panelID;
		
		// remove the old user-interface
		
		RemoveAppendedItems();
		
		// set up using the relevant mapper
		
		aMap = GetCurrentMapper();
		
		if ( aMap )
		{
			AppendItemsToDialog( aMap->GetDialogItemsID(), -kStdPrefUserPanelItem  );
			aMap->PanelEstablish();
		}
		
		// select the associated icon in the list
		
		ilb = dynamic_cast<ZIconListBox*>( GetObjectItem( kStdPrefSwitchIconList ));
		
		if ( ilb && panelID > 0 )
		{
			Cell c = { 0, 0 };
			
			c.v = panelID - 1;
			ilb->SelectCell( c );
		}
	}
}


ZPrefMapper*	ZStdPrefsDialog::GetCurrentMapper()
{
	if ( maps && curPanel > 0 )
		return maps->GetObject( curPanel );
	else
		return NULL;
}


void			ZStdPrefsDialog::GetPrefResInfo( short index, OSType* pType, short* pID, OSType* mapClass )
{
	ZPrefMapper*	aMap = maps->GetObject( index );
	
	if ( aMap )
	{		
		*pType = aMap->GetMapResourceType();
		*pID   = aMap->GetMapResourceID();
		
		if ( mapClass )
			*mapClass = aMap->GetMapClass();
			
	}
}


void			ZStdPrefsDialog::ProcessClickInIconList()
{
	ZIconListBox*	ilb = dynamic_cast<ZIconListBox*>( GetObjectItem( kStdPrefSwitchIconList ));	

	if ( ilb )
	{
		Cell c = { 0, 0 };
		
		ilb->GetSelectedCell( &c );
	
		SwitchPanels( c.v + 1 );
	}
}


void			ZStdPrefsDialog::SaveChangedPrefData()
{
	short			i, iMax = maps->CountItems();
	ZPrefMapper*	aMap;
	
	if ( gPrefsFile == NULL )
	{
		// no prefs file, so make sure we require one and create it
	
		FailNIL( gPrefsFile = new ZPrefsFile());
		gPrefsFile->OpenResFork();
	}
	
	for( i = 1; i <= iMax; i++ )
	{
		aMap = maps->GetObject( i );
		
		if ( aMap && aMap->DataChanged())
		{
			Handle		pData;
			
			// when saving, we always create a copy of the data and save that. The reason is
			// that the mapper always disposes the handle, since it is unaware whether the
			// dialog was cancelled or OK'd, and anyway, only changed items are written back,
			// so to prevent memory leaks, this approach is needed. For the same reason,
			// the original resource is detached when setting up the dialog.
			
			pData = aMap->GetPrefData();
			FailOSErr( HandToHand( &pData ));
		
			gPrefsFile->WriteResource( pData, aMap->GetMapResourceType(), aMap->GetMapResourceID());
			
			// inform mapper that we've saved its data. By default this broadcasts a notification
			// to its listeners. They can take action to read the new data & do whatever's appropriate.
			
			aMap->PrefChanged();
		}
	}

	gPrefsFile->Flush();
}


Boolean			ZStdPrefsDialog::DataToSave()
{
	short			i, iMax = maps->CountItems();
	ZPrefMapper*	aMap;
	Boolean			dSave = FALSE;
	
	for( i = 1; i <= iMax; i++ )
	{
		aMap = maps->GetObject( i );
		
		if ( aMap && aMap->DataChanged())
		{
			dSave = TRUE;
			break;
		}
	}
	
	return dSave;
}


#pragma mark -
#pragma mark ***** ZPrefMapper *****


ZPrefMapper::ZPrefMapper()
	: ZComrade()
{
	itsData = NULL;
	origDataCopy = NULL;
	pDialog = NULL;
	itemOffset = 0;
	dataChanged = FALSE;
	mapperClass = kStdMapperClassID;
	mapToPrefResType = '????';
	mapToPrefResID = 128;
	ditlID = -1;
}


ZPrefMapper::~ZPrefMapper()
{
	if ( itsData )
		DisposeHandle( itsData );
		
	if ( origDataCopy )
		DisposeHandle( origDataCopy );
}


void	ZPrefMapper::InitMapper( ZDialog* prefDialog,
								 short baseItems,
								 Handle prefData,
								 OSType classID,
								 OSType pResType,
							     short  pResID,
								 short dItemsID )
{
	itsData = prefData;
	pDialog = prefDialog;
	itemOffset = baseItems;
	dataChanged = FALSE;
	mapperClass = classID;
	mapToPrefResType = pResType;
	mapToPrefResID = pResID;
	ditlID = dItemsID;
	
	// make a copy of the data so we can do comparisons later on.
	
	if ( itsData )
	{
		origDataCopy = itsData;
		FailOSErr( HandToHand( &origDataCopy ));
	}
}


Handle	ZPrefMapper::MakeDefaultPrefData()
{
	// the default mapper's data is an array of shorts, each of which stores the control value
	// of the panel's controls. Edit fields are ignored. This allows you a very basic mapper
	// which is more or less a 1:1 map between dialog and data. However, most real-world
	// applications will want to do something else, since this approach is particularly crude.
	// NOTE that at this point, the actual number of items cannot be determined programmatically,
	// since the relevant panel is not switched in, thus we allocate an initial handle for
	// 32 items, and rely on a later call to Panelestablish to set it to the right size.
	
	// initially all items are set to 0.
	
	FailNIL( itsData = NewHandleClear( sizeof( short ) * 32 ));
	
	origDataCopy = itsData;
	FailOSErr( HandToHand( &origDataCopy ));
	return itsData;
}


void	ZPrefMapper::PanelEstablish()
{
	if ( itsData && pDialog )
	{
		short		i, iMax;
		short**		d = (short**) itsData;
		
		iMax = CountDITL( pDialog->GetMacWindow());
		
		// if data handle is not correct size for the number of panel items,
		// resize it appropriately. This guards against the case where items are added
		// to the dialog panel, but the same resource is targetted.
		
		if ( GetHandleSize( itsData ) != sizeof( short ) * ( iMax - itemOffset ))
		{
			SetHandleSize( itsData, sizeof( short ) * ( iMax - itemOffset ));
			FailMemError();
			
			dataChanged = TRUE;
		}
		
		// set dialog item controls to match the data
	
		for( i = itemOffset + 1; i <= iMax; i++ )
		{
			if ( pDialog->GetItemType( i ) & ctrlItem )
				pDialog->SetValue( i, (*d)[ i - itemOffset - 1 ] );
		}
	}
}


void	ZPrefMapper::PanelClick( short item )
{
	// the mapper can either keep the state of <itsData> handle matching the controls after
	// every click (as this does), or set it once in the PanelTearDown() method. The advantage
	// of doing it here is the ability to maintain the <dataChanged> flag accurately which the
	// owning dialog uses to enable or disable the "Save" button. This works here by directly
	// comparing the new data with the original one. Your own mappers can use other techniques
	// or something similar- whatever makes the most sense.
	
	if ( itsData && pDialog )
	{
		short	i, iMax;
		short**	d = (short**) itsData;
		
		iMax = CountDITL( pDialog->GetMacWindow());
	
		for( i = itemOffset + 1; i <= iMax; i++ )
		{
			if ( pDialog->GetItemType( i ) & ctrlItem )
				(*d)[ i - itemOffset - 1 ] = pDialog->GetValue( i );
		}

		dataChanged = ! EqualHandle( itsData, origDataCopy );
	}
}


void	ZPrefMapper::PrefChanged()
{
	SendMessage( msgPrefResourceHasChanged, (void*) mapperClass );
}


#pragma mark -

void	RegisterPrefMapper( ConstructorFunction	cf, OSType mapperClassID )
{
	FailNILParam( cf );

	if ( gPrefMapRegistry == NULL )
		gPrefMapRegistry = new ZArray( sizeof( MapInstRec ));
		
	long		i;
	MapInstRec	mi;
	
	for( i = 1; i<= gPrefMapRegistry->CountItems(); i++ )
	{
		gPrefMapRegistry->GetArrayItem( &mi, i );
	
		if ( mi.mapClassID == mapperClassID )
			return;
	}

	// here, passed map ID is unique
	
	mi.inst = cf;
	mi.mapClassID = mapperClassID;
	
	gPrefMapRegistry->AppendItem( &mi );
}