/*************************************************************************************************
*
*
*			MacZoop - "the framework for the rest of us"		 
*
*
*
*			ZStdPrefsDialog.h	-- a standard multi-panel preferences dialog.
*
*
*			 1997, Graham Cox
*
*
*
*************************************************************************************************/

#pragma once

#ifndef __ZSTDPREFSDIALOG__
#define	__ZSTDPREFSDIALOG__

#include	"ZAdvancedDialog.h"
#include	"ZExtraDialogItems.h"
#include	"ZObjectArray.h"

class	ZPrefsFile;

#if PRAGMA_ALIGN_SUPPORTED
#pragma options align=mac68k
#endif

// this structure provides info about each panel:

typedef struct
{
	short			iconID;					// res ID of 'cicn' resource for the panel
	short			ditlID;					// res ID of the 'DITL' resource that provides UI
	unsigned char	flags;					// flags for this panel
	char			spare;					// reserved- do not use
	short			prefResID;				// the ID of the pref resource this maps to
	OSType			prefResType;			// the type of the pref resource this maps to
	OSType			prefMapperClassID;		// the class ID of the mapper object
	Str31			prefPanelName;			// the name of the panel
}
PrefPanelRec;

// this structure is the resource that describes the entire dialog:

typedef struct
{
	short			version;				// version number of the resource format (1)
	short			relItemID;				// item ID of item panels are positioned by
	short			prefFlags;				// flags
	short			numPanels;				// number of specified panels to follow
	PrefPanelRec	pp[1];					// array of panel/mapper info
}
PrefDialogRec, *PrefDialogPtr, **PrefDialogHdl;

// dialog itself stores some prefs which make it nicer for the user

typedef struct
{
	short			version;				// version of this resource (1)
	short			openPanelIndex;			// index of open panel if class is not unique
	OSType			openPanelClass;			// class ID of the current panel's mapper
	Point			windowPos;				// global top,left of window
}
PrefDPrefRec, *PrefDPrefPtr, **PrefDPrefHdl;	

#define		kPrefDialogPrefResType		'PDP?'

#if PRAGMA_ALIGN_SUPPORTED
#pragma options align=mac68k
#endif

#define		kPrefDialogDescResType		'PINF'

typedef struct
{
	OSType				mapClassID;
	ConstructorFunction	inst;
}
MapInstRec;

/*

These classes and structures provide a sophisticated preferences dialog system. The dialog can
have multiple panels switched by an icon list, or a single panel if preferred. The dialog will
provide all of the management of the actual prefs data in the prefs file, all you need to do
is to provide a handler object that maps your dialog controls to the data format of the pref
resource.

The pref file contains a number of resources, each representing your own private information
that you wish to store- for example a set of flags, or names of things, etc. Anything you like.

The pref dialog contains controls which provide a user-interface to your pref data. If your
pref data is not editable, then you don't need to use any of this. If it is editable however,
you can provide a powerful and extensible standard interface with a few lines of code using
this.

Since your data format is private, you need to provide a pref mapping object that
maps dialog controls, etc to the data in the resource. The pref dialog will instantiate your
mapping objects as needed and call them to set up the dialog's controls, and to read the
dialog control states when the panel is closed. All the rest- saving the resource to the prefs
file, building and handling the dialog is all done for you.

*/


// the mapper class. In general you'll subclass this. The base mapper class will provide a
// very simple mapper that maps controls to an array of shorts. For some purposes, this will
// suffice, but you can achieve far more by implementing your own mappers. 

class	ZPrefMapper	: public ZComrade
{
protected:
	OSType		mapperClass;
	OSType		mapToPrefResType;
	Handle		itsData;
	Handle		origDataCopy;
	ZDialog*	pDialog;
	short		mapToPrefResID;
	short		itemOffset;
	short		ditlID;
	Boolean		dataChanged;
	
public:

	ZPrefMapper();
	virtual ~ZPrefMapper();
	
	virtual void	InitMapper( ZDialog* prefDialog,
								short baseItems,
								Handle prefData,
								OSType classID,
								OSType pResType,
								short pResID,
								short dItemsID );
	
	virtual Handle	MakeDefaultPrefData();
	virtual void	PanelEstablish();
	virtual void	PanelTearDown() {};
	virtual void	PanelClick( short item );
	virtual void	PrefChanged();
	virtual Handle	GetPrefData() { return itsData; };
	
	inline	Boolean	DataChanged() { return dataChanged; };
	inline	OSType	GetMapClass() { return mapperClass; };
	inline	OSType	GetMapResourceType() { return mapToPrefResType; };
	inline 	short	GetMapResourceID() { return mapToPrefResID; };
	inline  short	GetDialogItemsID() { return ditlID; };
};

typedef ZObjectArray<ZPrefMapper> ZPrefMapperList;

#define		kStdMapperClassID	'zpmp'

// this message transmitted by mapper when its pref data changed & saved to file:

enum
{
	msgPrefResourceHasChanged = 'prfx'
};


/*

the methods of a mapper are called in the following situations:

The mapper is instantiated when the dialog is instantiated.
The mapper is destroyed when the dialog is destroyed.

InitMapper is the first call you'll get. The dialog is fully built, but your panel is
not necessarily switched in. If the pref data specified in the PINF resource exists in the
pref file, its handle is passed to you. Otherwise you'll get NULL. Make a note of the dialog
ref and the baseItems count. InitMapper is called only once per prefs dialog session.

MakeDefaultPrefData() will be the next call, but only if no data was found in the file. In this
case, make a new handle of the appropriate size using NewHandle() and initialise its contents
to your default pref settings. Keep a reference to the handle in the mapper. Note that this
data does not belong to the mapper- you must never dispose it, though it is permissable to
resize it.

PanelEstablish() is called when the user activates your panel. You should read the state of
the data in your handle, and set the panel's controls, etc as appropriate to match. This may
be called more than once per prefs session.

PanelTearDown() is called when your panel is switched out in favour of another, or when the
Save() button is hit when your panel is still active. You must read the state of the dialog
items and set up the contents of your data handle appropriately. PanelEstablish() and
PanelTearDown() are always called in pairs.

PanelClick() is called when the user clicks items in your panel. You can take action as
necessary. For most controls such as checkboxes, radio buttons, plain buttons and edit text,
you don't need to do anything. If you wish to provide a more complex interface, you can use
this method to change the standard behaviour. Note that when dealing with dialog items, you
must add or subtract the base count as necessary to get the true ID of the item, since your
panel's items are appended to the dialog's base items. Do not assume the value of <baseItems>,
always use the value you're given.

PrefChanged() is called after the Save() button was clicked, and your data was changed (if
your data wasn't changed, you won't be called). This is just to inform you that your prefs
state was changed- you may want to inform other parts of your application to re-read the new
state and take appropriate action.

Note that the DITL panel you provide can have the full range of abilities you get with any
ZAdvancedDialog DITL, so often you won't need to do very much in your mapper to provide
sophisticated interfaces.

*/


// the dialog class:

class	ZStdPrefsDialog	: public ZAdvancedDialog
{
protected:
	ZPrefMapperList*	maps;
	short				curPanel;

public:
	ZStdPrefsDialog( ZCommander* aBoss, const short dialogID );
	virtual ~ZStdPrefsDialog();
	
	virtual void		SetUp();
	virtual void		ClickItem( const short item );
	virtual Boolean		CloseDialog();
	virtual void		ReceiveMessage( ZComrade* aSender, long msg, void* msgData );
	
protected:
	virtual ZPrefMapper*	MakeMapper( OSType classID );
	void					ParseDescriptorResource( const short resID );
	void					SwitchPanels( const short panelID );
	ZPrefMapper*			GetCurrentMapper();
	void					GetPrefResInfo( short index, OSType* pType, short* pID, OSType* mapClass = NULL );
	void					ProcessClickInIconList();
	void					SaveChangedPrefData();
	Boolean					DataToSave();
};

// std base items:

#define		kStdPrefSaveButton		1
#define		kStdPrefCancelButton	2
#define		kStdPrefSwitchIconList	3
#define		kStdPrefUserPanelItem	4

#define		kStdPrefsDialogResID	298

// your application can register mappers it defines by calling the following global function:

void	RegisterPrefMapper( ConstructorFunction cf, OSType mapperClassID );

// you need to establish a constructor function for each mapper you define, which simply
// makes a new mapper of the required type and returns it. The default mapper does not need
// to be registered. If you prefer, instead of registering a mapper, you can modify (or
// override) the method MakeMapper() above.

// Note that in order to provide some degree of protection for the future development of the
// framework, the following conventions apply:
// 
// Pref resource types consisting of all UPPER CASE letters are reserved by MacZoop. User
// classes and application code should use resource types containing at least one lower-case
// letter.
// 
// It is also advisable to publish the format of any pref resources you define if you make
// your code available to others, so that clashes of resource types are avoided. It is also
// good practice to take reasonable action to proof yourself against future expansion by
// declaring a set of "reserved" fields in your pref resource, and also establishing a
// version number field. Your mappers for such pref resources can be written to automatically
// convert pref data structures to later formats you define as a way to keep your users
// happy as your application lifecycle progresses.

#endif